#
# If you're reading this...
# Hi!  I'm Billy B. Bobby Bobertson Jr. III. -- my friend's call me Bob...
# I hope you are well, and are enjoying yourself wherever you may be. <3
#
# And If you are not reading this... ^^ this ^^ to you ALL the same (@everyone).
#
